
from pydevmgr_core import  NodeAlias, Defaults, NodeVar
from pydevmgr_core.base.class_recorder import record_class
from pydevmgr_vlt.base import VltDevice
from pydevmgr_vlt.devices._tools import _inc
from pydevmgr_vlt.devices.vltiodev.ctrl import COMMAND
from pydantic import create_model 

from enum import Enum
Base = VltDevice.Stat

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 
#                      _              _   
#   ___ ___  _ __  ___| |_ __ _ _ __ | |_ 
#  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
# | (_| (_) | | | \__ \ || (_| | | | | |_ 
#  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
# 

N_AI, N_DI, N_NI, N_TI  = [8]*4 



class STATUS(int, Enum):
    OK = 0
    ERROR = 1




    #  ____  _        _     ___       _             __                 
    # / ___|| |_ __ _| |_  |_ _|_ __ | |_ ___ _ __ / _| __ _  ___ ___  
    # \___ \| __/ _` | __|  | || '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \ 
    #  ___) | || (_| | |_   | || | | | ||  __/ |  |  _| (_| | (_|  __/ 
    # |____/ \__\__,_|\__| |___|_| |_|\__\___|_|  |_|  \__,_|\___\___| 


# some dinamicaly created nodes
io_nodes = {}
for i in range(N_DI):
    io_nodes[f'di_{i}'] = (ND, NC(suffix= f'stat.arr_DI[{i}].bValue'))
for i in range(N_AI):
    io_nodes[f'ai_{i}'] = (ND, NC(suffix= f'stat.arr_AI[{i}].lrValue'))
for i in range(N_NI):
    io_nodes[f'ni_{i}'] = (ND, NC(suffix= f'stat.arr_NI[{i}].nValue'))
for i in range(N_TI):
    io_nodes[f'ti_{i}'] = (ND, NC(suffix= f'stat.arr_TI[{i}].sValue'))



class VltIoDevStat(Base):
    STATUS = STATUS
    # Add the constants to this class 
    class Config( create_model("Config",  __base__ = Base.Config, **io_nodes)):
        initialised: ND = NC( suffix= 'stat.bInitialised' )
        last_command: ND = NC( suffix= 'stat.nLastCommand' )
        error_code: ND = NC( suffix= 'stat.nErrorCode' )
        error_text: ND = NC( suffix= 'stat.sErrorText' )
        status: ND = NC( suffix= 'stat.nStatus' )
    
    @NodeAlias.prop( nodes=[f'di_{i}' for i in range(N_DI) ] )
    def di_all(self, *flags):
        return flags
    @NodeAlias.prop( nodes=[f'ai_{i}' for i in range(N_AI)] )
    def ai_all(self, *values):
        return values
    @NodeAlias.prop( nodes=[f'ni_{i}' for i in range(N_NI)] )
    def ni_all(self, *values):
        return values
    @NodeAlias.prop( nodes=[f'ti_{i}' for i in range(N_TI)] )
    def ti_all(self, *values):
        return values

       
    # We can add some nodealias to compute some stuff on the fly 
    # If they node to be configured one can set a configuration above 
    
    # Node Alias here     
    # Build the Data object to be use with DataLink, the type and default are added here 
    class Data(Base.Data):
        initialised : NodeVar[bool] = False
        last_command: NodeVar[COMMAND] = COMMAND.NONE
        error_code: NodeVar[int] = 0 
        error_text: NodeVar[str] = ""
        status: NodeVar[STATUS] = STATUS.OK
        di_all: NodeVar[list] = []
        ai_all: NodeVar[list] = []
        ni_all: NodeVar[list] = []
        ti_all: NodeVar[list] = []



if __name__ == "__main__":
    VltIoDevStat( )



from pydevmgr_vlt.base import VltDevice
from pydevmgr_core import Defaults, NodeVar, NodeAlias, NodeAlias1
from pydevmgr_core.nodes import Opposite
from enum import Enum 
from typing import Optional 
from pydevmgr_vlt.devices.vltmotor.positions import PositionsConfig

Base = VltDevice.Stat

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 




class STATE(int, Enum):
    IDLE = 20
    RESET_AXIS = 30
    SET_POS = 40
    INIT = 50
    SAVE_TO_NOVRAM = 55
    CLEAR_NOVRAM = 57
    MOVE_ABS = 60
    MOVE_OPTIMISED = 61
    MOVE_VEL = 70
    CTRL_BRAKE = 75
    STOP = 80




class VltMotorStat(Base):
    STATE = STATE 
    class Config(Base.Config):
        mot_positions: PositionsConfig = PositionsConfig()
        state:                  ND  =  NC(suffix='nState'                     )
        pos_target:             ND  =  NC(suffix='stat.lrPosTarget'           )
        pos_actual:             ND  =  NC(suffix='stat.lrPosActual'           )
        vel_actual:             ND  =  NC(suffix='stat.lrVelActual'           )
        vel_target:             ND  =  NC(suffix='stat.lrVelTarget'           )
        axis_status:            ND  =  NC(suffix='stat.nAxisStatus'           )
        backlash_step:          ND  =  NC(suffix='stat.nBacklashStep'         )
        last_command:           ND  =  NC(suffix='stat.nLastCommand'          )
        error_code:             ND  =  NC(suffix='stat.nErrorCode'            )
        error_text:             ND  =  NC(suffix='stat.sErrorText'            )
        init_step:              ND  =  NC(suffix='stat.nInitStep'             )
        init_action:            ND  =  NC(suffix='stat.nInitAction'           )
        info_data1:             ND  =  NC(suffix='stat.nInfoData1'            )
        info_data2:             ND  =  NC(suffix='stat.nInfoData2'            )
        local:                  ND  =  NC(suffix='stat.bLocal'                )
        enabled:                ND  =  NC(suffix='stat.bEnabled'              )
        initialised:            ND  =  NC(suffix='stat.bInitialised'          )
        ref_switch:             ND  =  NC(suffix='stat.bRefSwitch'            )
        at_max_position:        ND  =  NC(suffix='stat.bAtMaxPosition'        )
        at_min_position:        ND  =  NC(suffix='stat.bAtMinPosition'        )
        limit_switch_positive:  ND  =  NC(suffix='stat.bLimitSwitchPositive'  )
        limit_switch_negative:  ND  =  NC(suffix='stat.bLimitSwitchNegative'  )
        brake_active:           ND  =  NC(suffix='stat.bBrakeActive'          )
        max_position:           ND  =  NC(suffix='stat.lrMaxPositionValue'    )
        min_position:           ND  =  NC(suffix='stat.lrMinPositionValue'    )
        mode:                   ND  =  NC(suffix='stat.nMode'                 )
        axis_ready:             ND  =  NC(suffix='stat.bAxisReady'            )
        moving_abs:             ND  =  NC(suffix='stat.bMovingAbs'            )
        moving_vel:             ND  =  NC(suffix='stat.bMovingVel'            )
        changing_vel:           ND  =  NC(suffix='stat.bChangingVel'          )
        

    @NodeAlias.prop("check", nodes=["error_code", "error_text"])
    def check(self, erc, ert):
        """ This node always return True but raise an error in case of device in error """
        if erc:
            raise RuntimeError(f"Error {erc}: {ert}")
        return True

    @NodeAlias1.prop( node="state")
    def state_txt(self, state):
        return self.STATE(state).name 

    @NodeAlias.prop( nodes=["state", "check"])
    def movement_finished(self, state, c):
        return state not in [self.STATE.MOVE_ABS, self.STATE.MOVE_OPTIMISED, self.STATE.MOVE_VEL, self.STATE.INIT]

    @NodeAlias.prop( nodes=["initialised", "check"])
    def initialisation_finished(self, initialised, c):
        return initialised

    @NodeAlias.prop( nodes=["enabled", "check"])
    def enable_finished(self, enabled, c):
        return enabled
    
    not_initialised = Opposite.prop(node="initialised")

    @NodeAlias1.prop(node="pos_actual")
    def pos_name(self, pos_actual):
        if not self.config.mot_positions: return ''
        positions = self.config.mot_positions
        tol = positions.tolerance
        for pname, pos in positions.positions.items():
            if abs( pos-pos_actual)<tol:
                return pname
        return ''

    # for this one we redefine the init so it does accept a mot_positions argument
    def __init__(self, *args, mot_positions: Optional[dict] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.config.mot_positions = mot_positions or PositionsConfig()

        
    # we need the mot_position from te parent (a Motor Device)
    # just add it to the dictionary create by  the super
    @classmethod
    def new_args(cls, parent, name, config):
        d = super().new_args(parent, name, config)
        try:
            mot_positions = parent.config.positions
        except AttributeError:
            mot_positions = {}
        d['mot_positions'] = mot_positions
        return d


    class Data(Base.Data):
        state:                  NV[int]    =  0
        pos_target:             NV[float]  =  0.0
        pos_actual:             NV[float]  =  0.0
        vel_actual:             NV[float]  =  0.0
        vel_target:             NV[float]  =  0.0
        axis_status:            NV[int]    =  0
        backlash_step:          NV[int]    =  0
        last_command:           NV[int]    =  0
        error_code:             NV[int]    =  0
        error_text:             NV[str]    =  ""
        init_step:              NV[int]    =  0
        init_action:            NV[int]    =  0
        info_data1:             NV[int]    =  0
        info_data2:             NV[int]    =  0
        local:                  NV[bool]   =  False
        enabled:                NV[bool]   =  False
        initialised:            NV[bool]   =  False
        ref_switch:             NV[bool]   =  False
        at_max_position:        NV[bool]   =  False
        at_min_position:        NV[bool]   =  False
        limit_switch_positive:  NV[bool]   =  False
        limit_switch_negative:  NV[bool]   =  False
        brake_active:           NV[bool]   =  False
        max_position:           NV[float]  =  0.0
        min_position:           NV[float]  =  0.0
        mode:                   NV[int]    =  0
        axis_ready:             NV[bool]   =  False
        moving_abs:             NV[bool]   =  False
        moving_vel:             NV[bool]   =  False
        changing_vel:           NV[bool]   =  False
           


if __name__ == "__main__":
    VltMotorStat()




from pydevmgr_core import  NodeAlias1, Defaults, NodeVar, record_class, BaseParser
from pydevmgr_vlt.base import VltDevice
from pydevmgr_vlt.devices._tools import _inc
from pydevmgr_ua import Int32

from enum import Enum
Base = VltDevice.Cfg

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 


class VltioDevCfg(Base):
    class Config(Base.Config):
        pass    
    class Data(Base.Data):
        pass           




from pydevmgr_vlt.base import VltDevice
from pydevmgr_core import Defaults, NodeVar, NodeAlias, NodeAlias1, record_class, BaseParser
from pydevmgr_ua import Int32, Int16
from enum import Enum 
from typing import Optional 
from pydevmgr_vlt.devices.vltmotor.positions import PositionsConfig
from pydevmgr_vlt.devices._tools import _inc 

Base = VltDevice.Ctrl

N = Base.Node # Base Node
NC = N.Config
ND = Defaults[NC] # this typing var says that it is a Node object holding default values 
NV = NodeVar # used in Data 


class MOTOR_COMMAND(int, Enum):
    NONE = _inc(0)
    INITIALISE = _inc()
    SET_POSITION = _inc()
    MOVE_ABSOLUTE = _inc()
    MOVE_RELATIVE = _inc()
    MOVE_VELOCITY = _inc()
    NEW_VELOCITY = _inc()
    NEW_POSITION = _inc()
    CLEAR_NOVRAM = _inc()

@record_class
class MotorCommand(BaseParser):
    class Config(BaseParser.Config):
        type = "MotorCommand"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(MOTOR_COMMAND, value)
        return Int32(value)


class DIRECTION(int, Enum):
    POSITIVE = _inc(1)
    SHORTEST = _inc()
    NEGATIVE = _inc()
    CURRENT  = _inc()

@record_class
class Direction(BaseParser):
    class Config(BaseParser.Config):
        type = "Direction"
    @staticmethod
    def parse(value, config):
        if isinstance(value, str):
            value =  getattr(DIRECTION, value)
        return Int16(value)





class VltMotorCtrl(Base):
    COMMAND = MOTOR_COMMAND
    DIRECTION = DIRECTION
    
    class Config(Base.Config):
        command:    ND  =  NC(  suffix=  'ctrl.nCommand',     parser=  MotorCommand  )
        direction:  ND  =  NC(  suffix=  'ctrl.nDirection',   parser=  Direction     )
        position:   ND  =  NC(  suffix=  'ctrl.lrPosition',   parser=  'UaDouble'    )
        velocity:   ND  =  NC(  suffix=  'ctrl.lrVelocity',   parser=  'UaDouble'    )
        stop:       ND  =  NC(  suffix=  'ctrl.bStop',        parser=  bool          )
        reset:      ND  =  NC(  suffix=  'ctrl.bResetError',  parser=  bool          )
        disable:    ND  =  NC(  suffix=  'ctrl.bDisable',     parser=  bool          )
        enable:     ND  =  NC(  suffix=  'ctrl.bEnable',      parser=  bool          )
        execute:    ND  =  NC(  suffix=  'ctrl.bExecute',     parser=  bool          )

    class Data(Base.Data):
        pass # empty for ctr TODO: add data for ctrl ? 
   

if __name__=="__main__":
    VltMotorCtrl()



from pydevmgr_vlt.devices.vltmotor.stat import VltMotorStat as Stat
from pydevmgr_vlt.devices.vltmotor.cfg import VltMotorCfg as Cfg, AXIS_TYPE
from pydevmgr_vlt.devices.vltmotor.ctrl import VltMotorCtrl as Ctrl
from pydevmgr_vlt.devices.vltmotor.init_seq import INITSEQ, InitialisationConfig, init_sequence_to_cfg
from pydevmgr_vlt.devices.vltmotor.positions import  PositionsConfig


from pydevmgr_vlt.base import VltDevice
from pydevmgr_core import upload 
from typing import Optional , Union , Dict, Any
from pydantic import BaseModel, validator

Base = VltDevice



class VltMotorCtrlConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    velocity : float = 0.1 # mendatory because used as default for movement
    min_pos :           Optional[float] = 0.0
    max_pos :           Optional[float] = 0.0 
    axis_type :         Union[None,int,str] = "LINEAR" # LINEAR , CIRCULAR, CIRCULAR_OPTIMISED
    active_low_lstop :  Optional[bool] = False
    active_low_lhw :    Optional[bool] = False
    active_low_ref :    Optional[bool] = True
    active_low_index :  Optional[bool] = False
    active_low_uhw :    Optional[bool] = True
    active_low_ustop :  Optional[bool] = False
    brake :             Optional[bool] = False
    low_brake :         Optional[bool] = False
    backlash :          Optional[float] = 0.0
    tout_init :         Optional[int] = 30000
    tout_move :         Optional[int] = 12000
    tout_switch :       Optional[int] = 10000
    
    scale_factor : Optional[float] = 1.0
    accel: Optional[float] = 30.0
    decel: Optional[float] = 30.0
    jerk:  Optional[float] = 100.0
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
    @validator('axis_type')
    def validate_axis_type(cls, ax):
        if isinstance(ax, str):
            try:
                getattr(AXIS_TYPE, ax)
            except AttributeError:
                raise ValueError(f"Unknown axis_type {ax!r}")
        if isinstance(ax, int):            
            # always return a string??
            ax = AXIS_TYPE(ax).name        
        return ax


class VltMotorConfig(Base.Config):
    CtrlConfig = VltMotorCtrlConfig
    Positions = PositionsConfig
    Initialisation = InitialisationConfig
    
    Ctrl = Ctrl.Config
    Stat = Stat.Config 
    Cfg = Cfg.Config

    ctrl : Ctrl = Ctrl()
    stat: Stat = Stat()
    cfg: Cfg = Cfg()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    type: str = "VltMotor"
    initialisation : Initialisation= Initialisation()
    positions      : Positions= Positions()
    ctrl_config    : CtrlConfig = CtrlConfig()



class VltMotor(Base):
    Config = VltMotorConfig
    

    Cfg  = Cfg 
    Stat = Stat
    Ctrl = Ctrl
    
    class Data(Base.Data):
        Stat = Stat.Data
        Cfg = Cfg.Data
        Ctrl = Ctrl.Data
        
        stat: Stat = Stat()
        cfg: Cfg = Cfg()
        ctrl: Ctrl = Ctrl()
        

    def get_configuration(self, exclude_unset=True,  **kwargs) -> Dict[VltDevice.Node,Any]:
        """  return a node/value pair dictionary ready to be uploaded 
        
        The node/value dictionary represent the device configuration. 
        
        Args:
            **kwargs : name/value pairs pointing to cfg.name node
                      This allow to change configuration on the fly
                      without changing the config file. 
        """
        
        config = self._config 
        
        ctrl_config = config.ctrl_config
        # just update what is in ctrl_config, this should work for motor 
        # one may need to check parse some variable more carefully       
        values = ctrl_config.dict(exclude_none=True, exclude_unset=exclude_unset)
        cfg_dict = {getattr(self.cfg, k):v for k,v in  values.items() }
        cfg_dict[self.is_ignored] = self.config.ignored 
        cfg_dict.update({getattr(self.cfg, k):v for k,v in  kwargs.items() })
        
        init_cfg = init_sequence_to_cfg(config.initialisation, self.Cfg.INITSEQ)
        cfg_dict.update({getattr( self.cfg, k):v for k,v in init_cfg.items()})
        

        AXIS_TYPE = self.Cfg.AXIS_TYPE
        
        # transform axis type to number 
        if self.cfg.axis_type in cfg_dict:
            axis_type = cfg_dict[self.cfg.axis_type] 
            cfg_dict[self.cfg.axis_type] =  getattr(AXIS_TYPE, axis_type) if isinstance(axis_type, str) else axis_type
        ###
        # Set the new config value to the device 
        return cfg_dict      
    
    @property
    def velocity(self) -> float:
        return self.config.ctrl_config.velocity
   

    def check(self):
        """Check if the device is in error. Raise an error in case it is True """
        return self.stat.check.get()

    def stop(self):
        self.ctrl.stop.set(True)
    
    def reset(self):
        self.ctrl.reset.set(True)
        return self.stat.not_initialised
    
    def enable(self):
        self.ctrl.enable.set(True)
        return self.stat.enable_finished
    
    def disable(self):
        self.ctrl.disable.set(True)
        return self.stat.enabled


    def init(self):
        upload({ 
            self.ctrl.execute : True, 
            self.ctrl.command : self.Ctrl.COMMAND.INITIALISE
            })
        return self.stat.initialisation_finished

    def move_abs(self, pos, vel=None):
        
        vel = self.velocity if vel is None else vel
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.velocity: vel, 
            self.ctrl.command : self.Ctrl.COMMAND.MOVE_ABSOLUTE
            })
        return self.stat.movement_finished

    def move_rel(self, pos, vel=None):

        vel = self.velocity if vel is None else vel
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.velocity: vel, 
            self.ctrl.command : self.Ctrl.COMMAND.MOVE_RELATIVE
            })
        return self.stat.movement_finished  

    def move_vel(self, vel):
        direction = self.Ctrl.DIRECTION.POSITIVE if vel>0 else self.Ctrl.DIRECTION.NEGATIVE
        vel = abs(vel)
        upload({
            self.ctrl.execute : True, 
            self.ctrl.velocity: vel, 
            self.ctrl.direction: direction, 
            self.ctrl.command: self.Ctrl.COMMAND.MOVE_VELOCITY
            })
        return None
    
    def get_pos_target_of_name(self, name: str) -> float:
        """return the configured target position of a given pos name or raise error"""
        try:
            position = getattr(self.config.positions, name)
        except AttributeError:
            raise ValueError('unknown posname %r'%name)
        return position

    def get_name_of_pos(self, pos_actual: float) -> str:
        """ Retrun the name of a position from a position as input or ''
        
        Example:
            m.get_name_of( m.stat.pos_actual.get() )
        """
        positions = self.config.positions    
        tol = positions.tolerance
        
        for pname, pos in positions.positions.items():
            if abs( pos-pos_actual)<tol:
                return pname
        return ''

    def move_name(self, name, vel=None) -> VltDevice.Node:
        """ move motor to a named position 
        
        Args:
           name (str): named position
           vel (float):   target velocity for the movement
        """
        absPos = self.get_pos_target_of_name(name)
        return self.move_abs(absPos, vel)
        

    def set_pos(self, pos):
        """ Set the curent position value """
        upload({
            self.ctrl.execute : True, 
            self.ctrl.position: pos, 
            self.ctrl.command : self.Ctrl.COMMAND.SET_POSITION
            })



if __name__=="__main__":
    VltMotor()
    print("OK")



from pydantic import BaseModel, root_validator  
from typing import List, Dict
from collections import OrderedDict

class PositionsConfig(BaseModel):
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Structure 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    posnames : List = []
    tolerance: float = 1.0
    positions: Dict = OrderedDict()  # adding a dictionary for positions. Presfered than leaving it as extra 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Data Validator Functions
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    class Config:        
        extra = 'allow' # needed for the poses 
        validate_assignment = True
    @root_validator()
    def collect_positions(cls, values):     
        """ collectect the positions from the extras """ 
        positions = values['positions']
        for name in values['posnames']:
            if name not in positions:
                try:
                    positions[name] = float( values[name] ) 
                except (KeyError, TypeError):
                    raise ValueError(f'posname {name!r} is not defined or not a float')   
        return values 
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Method to save back the configuration     
    def cfgdict(self):
        d = {'posnames': self.posnames, 'tolerance':self.tolerance}
        for p in self.posnames:
            d[p] = self.positions[p]
        return d
# ################################  


